#define PYFT2_MODULE
#include "pyft2.h"

#define GLYPH_FUNC(s, c) PyCFunction_New(&Glyph_methods[c], (PyObject*)s)

static PyObject *
Glyph_transform(Glyph *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "xx", "xy", "yx", "yy","dx","dy", NULL };
	FT_Matrix m = { 1 << 16, 0, 0, 1 << 16};
	FT_Vector v = {0, 0};
	
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|iiiiii", kwlist,
					&m.xx, &m.xy, &m.yx, &m.yy,
					&v.x, &v.y))
		return NULL;
	
	if(FT_Glyph_Transform(self->glyph, &m, &v)) {
		PyErr_SetString(PyExc_SystemError,
				"FT_Glyph_Transform() failed");		
		return NULL;
	}
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Glyph_toBitmap(Glyph *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "mode", "dx", "dy", NULL };
	FT_Vector v = {0, 0};
	int mode;
	mode = FT_RENDER_MODE_NORMAL;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|iii", kwlist,
					&mode,&v.x, &v.y))
		return NULL;
	
	/* XXX : Do we always want to destroy the previous bitmap? */
	if(FT_Glyph_To_Bitmap(&self->glyph, mode, &v, (FT_Bool)1)) {
		PyErr_SetString(PyExc_SystemError,
				"FT_Glyph_To_Bitmap() failed");
		return NULL;
	}
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
Glyph_getCBox(Glyph *self, PyObject *args, PyObject *kwds) {
	static char *kwlist[]={ "mode", NULL };	
	int mode;
	FT_BBox bb;
	
	mode = FT_GLYPH_BBOX_PIXELS;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "|i", kwlist,&mode))
		return NULL;
	
	FT_Glyph_Get_CBox(self->glyph, (FT_UInt)mode, &bb);
	
	return Py_BuildValue("(iiii)", bb.xMin, bb.yMin, bb.xMax, bb.yMax);
}

static PyObject *
Glyph_copy(Glyph *self) {
	
	Glyph *glyph;
	
	glyph = (Glyph*)((GlyphType.tp_alloc)(&GlyphType, 0));
	if (glyph == NULL) return NULL;
	glyph->refface = self->refface;
	if(FT_Glyph_Copy(self->glyph, &glyph->glyph)) {
		PyErr_SetString(PyExc_SystemError, "FT_Render_Glyph failed");
		PyObject_Del(glyph);
		return NULL;
	}
	Py_XINCREF(self->refface);
	return (PyObject *)glyph;
}

enum {
	GLYPH_TRANSFORM, /* scalable */
	GLYPH_GETCBOX,   /* */
	GLYPH_TOBITMAP,  /* scalable */
	GLYPH_COPY
};

static struct PyMethodDef Glyph_methods[] = {
	{ "transform", (PyCFunction)Glyph_transform,   METH_KEYWORDS,
	  (char*)NULL  },
	{ "getCBox",   (PyCFunction)Glyph_getCBox,     METH_KEYWORDS,
	  (char*)NULL  },
	{ "toBitmap",  (PyCFunction)Glyph_toBitmap,    METH_KEYWORDS,
	  (char*)NULL  },
	{ "Glyph",     (PyCFunction)Glyph_copy,        METH_NOARGS,
	  (char*)NULL  },
	{ NULL, NULL, NULL, NULL }
};

static PyObject*
Glyph_getattr(Glyph *self, char *name) {
	switch (name[0]) {
	case 'G':
		if(!strcmp(name,"Glyph"))
			return GLYPH_FUNC(self, GLYPH_COPY);
		break;
	case 'b':
		if(self->glyph->format == FT_GLYPH_FORMAT_BITMAP)
		   if(!strcmp(name,"bitmap"))
		      return Py_BuildValue(
			      "iiii",
			      ((FT_BitmapGlyph)(self->glyph))->left,
			      ((FT_BitmapGlyph)(self->glyph))->top,
			      ((FT_BitmapGlyph)(self->glyph))->bitmap.width,
			      ((FT_BitmapGlyph)(self->glyph))->bitmap.rows);
		break;
	case 'f':
		if(!strcmp(name,"format"))
			return PyInt_FromLong((long)self->glyph->format);
		break;
	case 'g':
		if(!strcmp(name,"getCBox"))
			return GLYPH_FUNC(self, GLYPH_GETCBOX);
		break;
	case 't':
		if(!strcmp(name,"transform"))
			return GLYPH_FUNC(self, GLYPH_TRANSFORM);
		
		if(!strcmp(name,"toBitmap"))
			return GLYPH_FUNC(self, GLYPH_TOBITMAP);
		break;
	}
	
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}

static void Glyph_dealloc(Glyph *self) {
	printf("Glyph_dealloc() -> %p\n",self);
	FT_Done_Glyph(self->glyph);
	Py_DECREF((PyObject*)(self->refface));	
	PyObject_Del(self);
}

static PyObject*
Glyph_str(Glyph * self) {
	return PyString_FromFormat("<Glyph at %p>",self->glyph);
}

static PyObject*
Glyph_repr(Glyph * self) {
	return PyString_FromString("Glyph()");
}

PyTypeObject GlyphType = {
	PyObject_HEAD_INIT(NULL)
	0,                          /* ob_size         */
	"GGI._ft2module.Glyph",     /* tp_name         */
	sizeof(Glyph),              /* tp_basicsize    */
	0,                          /* tp_itemsize     */
	(destructor)Glyph_dealloc,  /* tp_dealloc      */
	0,                          /* tp_print        */
	(getattrfunc)Glyph_getattr, /* tp_getattr      */
	(setattrfunc)0, /* tp_setattr      */
	0,                          /* tp_compare      */
	(reprfunc)Glyph_repr,       /* tp_repr         */
	0,                          /* tp_as_number    */
	0,                          /* tp_as_sequence  */
	0,                          /* tp_as_mapping   */
	0,                          /* tp_hash         */
	0,                          /* tp_call         */
	(reprfunc)Glyph_str,        /* tp_str          */
	0,                          /* tp_getattro     */
	0,                          /* tp_setattro     */
	0,                          /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Glyph object",            /* tp_doc          */
	
	/* 2.0 */
	0,		            /* tp_traverse     */
	0,		            /* tp_clear        */
	/* 2.1 */
	0,		            /* tp_richcompare  */
	0,		            /* tp_weaklistoffset */
	/* 2.2 */
	0,		            /* tp_iter         */
	0,		            /* tp_iternext     */
	0,                          /* tp_methods      */
	0,                          /* tp_members      */
	0,                          /* tp_getset       */
	0,                          /* tp_base         */
	0,                          /* tp_dict         */
	0,                          /* tp_descr_get    */
	0,                          /* tp_descr_set    */
	0,                          /* tp_dictoffset   */
	(initproc)0,                /* tp_init         */
	0,                          /* tp_alloc        */
	0, /* can't be created except in Face */
	0,                          /* tp_is_gc        */
	0,                          /* tp_bases        */
        0,                          /* tp_mro          */
	0,                          /* tp_cache        */
	0,                          /* tp_subclasses   */
	0,                          /* tp_weaklist     */
	0                           /* tp_del          */
};
